"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import logging

from biweeklybudget.cliutils import (
    set_log_level_format, set_log_info, set_log_debug
)

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock  # noqa
else:
    from unittest.mock import patch, call, Mock  # noqa

pbm = 'biweeklybudget.cliutils'


class TestCliUtils(object):

    def test_set_log_info(self):
        mock_log = Mock(spec_set=logging.Logger)
        with patch('%s.set_log_level_format' % pbm) as mock_set:
            set_log_info(mock_log)
        assert mock_set.mock_calls == [
            call(
                mock_log, logging.INFO,
                '%(asctime)s %(levelname)s:%(name)s:%(message)s'
            )
        ]

    def test_set_log_debug(self):
        mock_log = Mock(spec_set=logging.Logger)
        with patch('%s.set_log_level_format' % pbm) as mock_set:
            set_log_debug(mock_log)
        assert mock_set.mock_calls == [
            call(mock_log, logging.DEBUG,
                 "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - "
                 "%(name)s.%(funcName)s() ] %(message)s")
        ]

    def test_set_log_level_format(self):
        mock_log = Mock(spec_set=logging.Logger)
        mock_handler = Mock(spec_set=logging.Handler)
        type(mock_log).handlers = [mock_handler]
        with patch('%s.logging.Formatter' % pbm) as mock_formatter:
            set_log_level_format(mock_log, 5, 'foo')
        assert mock_formatter.mock_calls == [
            call(fmt='foo')
        ]
        assert mock_handler.mock_calls == [
            call.setFormatter(mock_formatter.return_value)
        ]
        assert mock_log.mock_calls == [
            call.setLevel(5)
        ]



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
import codecs
import urllib
import json
from tempfile import mkstemp

from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

logger = logging.getLogger(__name__)


class ScreenScraper(object):
    """
    Base class for screen-scraping bank/financial websites.
    """

    def __init__(self, savedir='./', screenshot=False):
        """
        Initialize ScreenScraper.

        :param savedir: directory to save OFX in
        :type savedir: str
        :param screenshot: whether or not to take screenshots throughout the
          process
        :type screenshot: bool
        """
        self._savedir = os.path.abspath(os.path.expanduser(savedir))
        if not os.path.exists(self._savedir):
            os.makedirs(self._savedir)
        self._cookie_file = os.path.join(self._savedir, 'cookies.txt')
        logger.debug('Using savedir: %s', self._savedir)
        self._screenshot_num = 1
        self._screenshot = screenshot
        if self._screenshot:
            logger.warning("screenshotting all actions")
        self.user_agent = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36' \
                          ' (KHTML, like Gecko) Chrome/62.0.3202.62 ' \
                          'Safari/537.36'
        # temporary file for driver logs
        fp, self._service_log_path = mkstemp()
        os.close(fp)

    def __del__(self):
        try:
            os.unlink(self._service_log_path)
        except Exception:
            pass

    def load_cookies(self, cookie_file):
        """
        Load cookies from a JSON cookie file on disk. This file is not the
        format used natively by PhantomJS, but rather the JSON-serialized
        representation of the dict returned by
        :py:meth:`selenium.webdriver.remote.webdriver.WebDriver.get_cookies`.

        Cookies are loaded via
        :py:meth:`selenium.webdriver.remote.webdriver.WebDriver.add_cookie`

        :param cookie_file: path to the cookie file on disk
        :type cookie_file: str
        """
        if not os.path.exists(cookie_file):
            logger.warning('Cookie file does not exist: %s', cookie_file)
            return
        logger.info('Loading cookies from: %s', cookie_file)
        with open(cookie_file, 'r') as fh:
            cookies = json.loads(fh.read())
        count = 0
        for c in cookies:
            try:
                self.browser.add_cookie(c)
                count += 1
            except Exception as ex:
                logger.info('Error loading cookie %s: %s', c, ex)
        logger.debug('Loaded %d of %d cookies', count, len(cookies))

    def save_cookies(self, cookie_file):
        """
        Save cookies to a JSON cookie file on disk. This file is not the
        format used natively by PhantomJS, but rather the JSON-serialized
        representation of the dict returned by
        :py:meth:`selenium.webdriver.remote.webdriver.WebDriver.get_cookies`.

        :param cookie_file: path to the cookie file on disk
        :type cookie_file: str
        """
        cookies = self.browser.get_cookies()
        raw = json.dumps(cookies)
        with open(cookie_file, 'w') as fh:
            fh.write(raw)
        logger.info('Wrote %d cookies to: %s', len(cookies), cookie_file)

    def do_screenshot(self):
        """take a debug screenshot"""
        if not self._screenshot:
            return
        fname = os.path.join(
            os.getcwd(), '{n}.png'.format(n=self._screenshot_num)
        )
        self._pre_screenshot()
        self.browser.get_screenshot_as_file(fname)
        self._post_screenshot()
        logger.debug(
            "Screenshot: {f} of: {s}".format(
                f=fname,
                s=self.browser.current_url
            )
        )
        self._screenshot_num += 1

    def error_screenshot(self, fname=None):
        cwd = os.getcwd()
        if fname is None:
            fname = os.path.join(cwd, 'webdriver_fail.png')
        self._pre_screenshot()
        self.browser.get_screenshot_as_file(fname)
        self._post_screenshot()
        logger.error("Screenshot saved to: {s}".format(s=fname))
        logger.error("Page title: %s", self.browser.title)
        logger.error('Page URL: %s', self.browser.current_url)
        html_path = os.path.join(cwd, 'webdriver_fail.html')
        source = self.browser.execute_script(
            "return document.getElementsByTagName('html')[0].innerHTML"
        )
        with codecs.open(html_path, 'w', 'utf-8') as fh:
            fh.write(source)
        logger.error('Page source saved to: %s', html_path)
        if (
            os.path.exists(self._service_log_path) and
            os.path.getsize(self._service_log_path) > 2
        ):
            logpath = os.path.join(cwd, 'webdriver_service_log.txt')
            with open(logpath, 'w') as svclog:
                with open(self._service_log_path, 'r') as orig:
                    svclog.write(orig.read())
            logger.error('Webdriver driver log written to: %s', logpath)
        try:
            log_types = self.browser.log_types
        except Exception:
            logger.error('Failed to gather browser logs', excinfo=True)
            return
        for name in log_types:
            try:
                log = self.browser.get_log(name)
            except Exception:
                logger.error(
                    'Failed to get %s log from browser', name, excinfo=True
                )
                continue
            logpath = os.path.join(cwd, 'webdriver_log_%s.txt' % name)
            with open(logpath, 'w') as fh:
                if isinstance(log, type([])):
                    fh.write("\n".join([str(x) for x in log]))
                else:
                    fh.write(log)
            logger.error(
                'Wrote driver\'s "%s" log (length: %d) to: %s',
                name, len(log), logpath
            )

    def _pre_screenshot(self):
        if not self._browser_name.startswith('chrome'):
            return
        height = self.browser.execute_script(
            "return Math.max(document.body.scrollHeight, "
            "document.body.offsetHeight, document.documentElement."
            "clientHeight, document.documentElement.scrollHeight, "
            "document.documentElement.offsetHeight);"
        )
        height += 100
        logger.info('Resizing browser to %d high', height)
        self.browser.set_window_size(1920, height)

    def _post_screenshot(self):
        if not self._browser_name.startswith('chrome'):
            return
        self.browser.set_window_size(1920, 1080)

    def xhr_get_url(self, url):
        """ use JS to download a given URL, return its contents """
        script = 'var xhr = new XMLHttpRequest(); '
        script += 'var jantman_dl_response = null; '
        script += 'xhr.open("GET", "{url}", false); '.format(url=url)
        script += 'xhr.send(null); '
        script += 'jantman_dl_response = xhr.response; '
        script += 'return jantman_dl_response;'
        logger.debug("executing in browser: {s}".format(s=script))
        res = self.browser.execute_script(script)
        logger.debug("got {c} characters of return value from script".format(
            c=len(res)))
        return res

    def xhr_post_urlencoded(self, url, data, headers={}):
        """ use JS to download a given URL, return its contents """
        if not isinstance(data, type('')) and not isinstance(data, type(u'')):
            data = urllib.urlencode(data)
        script = 'var xhr = new XMLHttpRequest(); '
        script += 'var jantman_dl_response = null; '
        script += 'xhr.open("POST", "{url}", false); '.format(url=url)
        if isinstance(headers, type({})):
            headers["Content-type"] = "application/x-www-form-urlencoded"
            for k, v in headers.items():
                script += 'xhr.setRequestHeader("%s", "%s"); ' % (k, v)
        else:
            for item in headers:
                script += 'xhr.setRequestHeader("%s", "%s"); ' % (
                    item[0], item[1]
                )
        script += 'xhr.send("{p}"); '.format(p=data)
        script += 'jantman_dl_response = { resp: xhr.response, ' \
                  'respText: xhr.responseText, status: xhr.status, ' \
                  'headers: xhr.getAllResponseHeaders() }; '
        script += 'return JSON.stringify(jantman_dl_response);'
        logger.debug("executing in browser: %s", script)
        res = self.browser.execute_script(script)
        j = json.loads(res)
        logger.debug("Script returned %d length result (status %s; "
                     "headers: %s)", len(j['respText']),
                     j['status'],
                     j['headers'].replace("\r", "").replace("\n", "; "))
        return j['respText']

    def get_browser(self, browser_name, useragent=None):
        """
        get a webdriver browser instance

        :param browser_name: name of browser to get. Can be one of "firefox",
          "chrome", "chrome-headless", or "phantomjs"
        :type browser_name: str
        :param useragent: Optionally override the browser's default user-agent
          string with this value. Supported for phantomjs or chrome.
        :type useragent: str
        """
        self._browser_name = browser_name
        if browser_name == 'firefox':
            logger.debug("getting Firefox browser (local)")
            if 'DISPLAY' not in os.environ:
                logger.debug("exporting DISPLAY=:0")
                os.environ['DISPLAY'] = ":0"
            browser = webdriver.Firefox()
        elif browser_name in ['chrome', 'chrome-headless']:
            chrome_options = Options()
            if browser_name == 'chrome-headless':
                logger.debug('getting Chrome browser (local) with --headless')
                chrome_options.add_argument("--headless")
            else:
                logger.debug("getting Chrome browser (local)")
            if useragent is not None:
                chrome_options.add_argument('--user-agent=%s' % useragent)
                logger.debug(
                    'Setting chrome user-agent to "%s"', useragent
                )
            browser = webdriver.Chrome(
                chrome_options=chrome_options, desired_capabilities={
                    'loggingPrefs': {'browser': 'ALL'}
                },
                service_log_path=self._service_log_path
            )
            browser.set_window_size(1920, 1080)
            browser.implicitly_wait(2)
        elif browser_name == 'phantomjs':
            logger.debug("getting PhantomJS browser (local)")
            dcap = dict(DesiredCapabilities.PHANTOMJS)
            if useragent is None:
                logger.debug(
                    'Setting phantomjs user-agent to "%s"', self.user_agent
                )
                dcap["phantomjs.page.settings.userAgent"] = self.user_agent
            else:
                dcap["phantomjs.page.settings.userAgent"] = useragent
                logger.debug(
                    'Setting phantomjs user-agent to "%s"', useragent
                )
            args = [
                '--cookies-file={c}'.format(c=self._cookie_file),
                '--ssl-protocol=any',
                '--ignore-ssl-errors=true',
                '--web-security=false'
            ]
            browser = webdriver.PhantomJS(
                desired_capabilities=dcap, service_args=args
            )
            browser.set_window_size(1024, 768)
        else:
            raise SystemExit(
                "ERROR: browser type must be one of 'firefox', 'chrome', "
                "'chrome-headless' or 'phantomjs', not '{b}'".format(
                    b=browser_name
                )
            )
        logger.debug("returning browser")
        return browser

    def doc_readystate_is_complete(self, foo):
        """ return true if document is ready/complete, false otherwise """
        result_str = self.browser.execute_script("return document.readyState")
        if result_str == "complete":
            return True
        return False

    def jquery_finished(self, foo):
        """ return true if jQuery.active == 0 else false """
        active = self.browser.execute_script("return jQuery.active")
        if active == 0:
            return True
        return False

    def wait_for_ajax_load(self, timeout=20):
        """
        Function to wait for an ajax event to finish and trigger page load,
        like the Janrain login form.

        Pieced together from
        http://stackoverflow.com/a/15791319

        timeout is in seconds
        """
        WebDriverWait(self.browser, timeout).until(
            self.doc_readystate_is_complete
        )
        return True



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from time import sleep
from decimal import Decimal
from selenium.common.exceptions import (
    StaleElementReferenceException, TimeoutException, WebDriverException
)
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from biweeklybudget.utils import fmt_currency

logger = logging.getLogger(__name__)


class AcceptanceHelper(object):

    def normalize_html(self, html):
        """
        Given a HTML string, normalize some differences that may occur between
        different test environments.

        :param html: html
        :type html: str
        :return: normalized HTML
        :rtype: str
        """
        # strange inconsistency between local and CI...
        html = html.replace('style="display: none; "', 'style="display: none;"')
        return html

    def relurl(self, url):
        """
        Given an absolute URL including ``self.baseurl``, return the relative
        URL portion excluding the base url.

        :param url: base url
        :type url: str
        :returns: relative URL
        :rtype: str
        """
        return url.replace(self.baseurl, '')

    def thead2list(self, elem):
        """
        Given a webdriver table element, return the inner text strings of each
        ``th`` within the ``thead``, left to right.

        :param elem: table element
        :type elem: selenium.webdriver.remote.webelement.WebElement
        :return: list of table heading strings in order left to right
        :rtype: list
        """
        thead = elem.find_element(By.TAG_NAME, 'thead')
        tr = thead.find_element(By.TAG_NAME, 'tr')
        cells = []
        for th in tr.find_elements(By.TAG_NAME, 'th'):
            cells.append(th.text.strip())
        return cells

    def thead2elemlist(self, elem):
        """
        Given a webdriver table element, return the WebElements of each
        ``th`` within the ``thead``, left to right.

        :param elem: table element
        :type elem: selenium.webdriver.remote.webelement.WebElement
        :return: list of table heading WebElements in order left to right
        :rtype: list
        """
        thead = elem.find_element(By.TAG_NAME, 'thead')
        tr = thead.find_element(By.TAG_NAME, 'tr')
        cells = []
        for th in tr.find_elements(By.TAG_NAME, 'th'):
            cells.append(th)
        return cells

    def tbody2textlist(self, elem):
        """
        Given a webdriver ``table`` element, return a list of table rows, top to
        bottom, each being represented by a list of strings corresponding to
        the text content of each column in the row, left to right.

        :param elem: table element
        :type elem: selenium.webdriver.remote.webelement.WebElement
        :return: list of table rows, each being a list of cell content strings
        :rtype: list
        """
        tbody = elem.find_element(By.TAG_NAME, 'tbody')
        rows = []
        for tr in tbody.find_elements(By.TAG_NAME, 'tr'):
            row = []
            for td in tr.find_elements(By.XPATH, '*'):
                if td.tag_name not in ['td', 'th']:
                    continue
                row.append(td.text.strip())
            rows.append(row)
        return rows

    def tbody2trlist(self, elem):
        """
        Given a webdriver ``table`` element, return a list of table rows, top to
        bottom.

        :param elem: table element
        :type elem: selenium.webdriver.remote.webelement.WebElement
        :return: list of ``tr`` WebElements
        :rtype: list
        """
        tbody = elem.find_element(By.TAG_NAME, 'tbody')
        return [x for x in tbody.find_elements(By.TAG_NAME, 'tr')]

    def tbody2elemlist(self, elem):
        """
        Given a webdriver ``table`` element, return a list of table rows, top to
        bottom, each being represented by a list of WebElements corresponding to
        the cells of each column in the row (td or th), left to right.

        :param elem: table element
        :type elem: selenium.webdriver.remote.webelement.WebElement
        :return: list of table rows, each being a list of cell WebElements
        :rtype: list
        """
        tbody = elem.find_element(By.TAG_NAME, 'tbody')
        rows = []
        for tr in tbody.find_elements(By.TAG_NAME, 'tr'):
            row = []
            for td in tr.find_elements(By.XPATH, '*'):
                if td.tag_name not in ['td', 'th']:
                    continue
                row.append(td)
            rows.append(row)
        return rows

    def retry_stale(self, func, *args, **kwargs):
        """
        Retry calling ``func`` with ``*args, **kwargs`` up to 5 times, sleeping
        1 second between each, if a StaleElementReferenceException is found.
        Return its return value.
        """
        e = None
        for i in range(5):
            try:
                res = func(*args, **kwargs)
                return res
            except StaleElementReferenceException as ex:
                e = ex
                sleep(1)
        raise e

    def wait_for_modal_shown(self, driver):
        """
        Wait for the modal to be shown.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        """
        self.wait_until_clickable(driver, 'modalLabel', timeout=10)

    def wait_until_clickable(self, driver, elem_id, by=By.ID, timeout=10):
        """
        Wait for the modal to be shown.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param elem_id: element ID
        :type elem_id: str
        :param by: What method to use to find the element. This must be one of
          the strings which are values of
          :py:class:`selenium.webdriver.common.by.By` attributes.
        :type by: str
        :param timeout: timeout in seconds
        :type timeout: int
        """
        WebDriverWait(driver, timeout).until(
            EC.element_to_be_clickable((by, elem_id))
        )

    def wait_for_id(self, driver, id):
        """
        Wait for the element with ID ``id`` to be shown.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param id: ID of the element
        :type id: str
        """
        WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, id))
        )

    def wait_for_jquery_done(self, driver, sleep_sec=0.2, tries=20):
        """
        Wait until ``jQuery.active == 0``. Tries ``tries`` times at
        ``sleep`` second intervals; raises an exception if all tries fail.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param sleep_sec: how long to sleep between tries
        :type sleep_sec: float
        :param tries: how many times to try
        :type tries: bool
        """
        script = 'return jQuery.active == 0'
        count = 0
        while count < 20:
            res = driver.execute_script(script)
            if res:
                logger.debug(
                    'jQuery done after %d seconds', (sleep_sec * tries)
                )
                break
            sleep(sleep_sec)
        else:
            raise RuntimeError(
                'jQuery did not finish after %d seconds',
                (sleep_sec * tries)
            )

    def wait_for_datatable_rows(
        self, driver, table_id, expected_count, sleep_sec=0.5, tries=10
    ):
        """
        Wait until the DataTable with given ID has the expected number of rows.
        This is useful for waiting for DataTables search/filter to complete.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param table_id: The ID of the table element
        :type table_id: str
        :param expected_count: The expected number of rows
        :type expected_count: int
        :param sleep_sec: how long to sleep between tries
        :type sleep_sec: float
        :param tries: how many times to try
        :type tries: int
        """
        for count in range(tries):
            try:
                table = driver.find_element(By.ID, table_id)
                rows = table.find_element(
                    By.TAG_NAME, 'tbody'
                ).find_elements(By.TAG_NAME, 'tr')
                if len(rows) == expected_count:
                    logger.debug(
                        'DataTable %s has %d rows after %d tries',
                        table_id, expected_count, count + 1
                    )
                    return
            except StaleElementReferenceException:
                pass
            sleep(sleep_sec)
        raise RuntimeError(
            f'DataTable {table_id} did not reach {expected_count} rows after '
            f'{tries} tries ({sleep_sec * tries} seconds)'
        )

    def wait_for_load_complete(self, driver, sleep_sec=0.2, tries=20):
        """
        Wait until ``document.readyState == "complete"``. Tries ``tries`` times
        at ``sleep`` second intervals; raises an exception if all tries fail.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param sleep_sec: how long to sleep between tries
        :type sleep_sec: float
        :param tries: how many times to try
        :type tries: bool
        """
        script = 'return document.readyState == "complete"'
        count = 0
        while count < 20:
            res = driver.execute_script(script)
            if res:
                logger.debug(
                    'readyState complete after %d seconds', (sleep_sec * tries)
                )
                break
            sleep(sleep_sec)
        else:
            raise RuntimeError(
                'readyState did not reach complete after %d seconds',
                (sleep_sec * tries)
            )

    def get_modal_parts(self, selenium, wait=True):
        """
        Return a 3-tuple of the WebElements representing the modalDiv,
        modalLabel h4 and modalBody div.

        :param selenium: Selenium driver instance
        :type selenium: selenium.webdriver.remote.webdriver.WebDriver
        :param wait: whether or not to wait for presence of modalLabel
        :type wait: bool
        :return: 3-tuple of (modalDiv WebElement, modalLabel WebElement,
          modalBody WebElement)
        :rtype: tuple
        """
        count = 0
        while True:
            count += 1
            try:
                if wait:
                    self.wait_for_modal_shown(selenium)
                modal = selenium.find_element(By.ID, 'modalDiv')
                title = selenium.find_element(By.ID, 'modalLabel')
                body = selenium.find_element(By.ID, 'modalBody')
                return modal, title, body
            except TimeoutException:
                if count > 6:
                    raise
                print(
                    'TimeoutException waiting for modal to be shown; '
                    'try again in 3 seconds.'
                )
                sleep(3)
            except Exception:
                raise
        return None, None, None

    def assert_modal_displayed(self, modal, title, body):
        """
        Assert that the modal is displayed.

        :param modal: the modal itself
        :type modal: selenium.webdriver.remote.webelement.WebElement
        :param title: the title element of the modal
        :type title: selenium.webdriver.remote.webelement.WebElement
        :param body: the body element of the modal
        :type body: selenium.webdriver.remote.webelement.WebElement
        """
        assert modal.is_displayed()
        assert modal.is_enabled()
        assert title.is_displayed()
        assert title.is_enabled()
        assert body.is_displayed()
        assert body.is_enabled()

    def assert_modal_hidden(self, modal, title, body):
        """
        Assert that the modal is displayed.

        :param modal: the modal itself
        :type modal: selenium.webdriver.remote.webelement.WebElement
        :param title: the title element of the modal
        :type title: selenium.webdriver.remote.webelement.WebElement
        :param body: the body element of the modal
        :type body: selenium.webdriver.remote.webelement.WebElement
        """
        assert modal.is_displayed() is False
        assert title.is_displayed() is False
        assert body.is_displayed() is False

    def get(self, _selenium, url):
        """
        Wrapper around ``selenium`` fixture's ``get()`` method to retry up to
        4 times on TimeoutException.

        :param _selenium: selenium fixture instance
        :param url: URL to get
        :type url: str
        """
        count = 0
        while True:
            count += 1
            try:
                _selenium.get(url)
                return
            except TimeoutException:
                if count > 4:
                    raise
                print('selenium.get(%s) timed out; trying again', url)
            except Exception:
                raise
        self.wait_for_load_complete(_selenium)
        self.wait_for_jquery_done(_selenium)

    def inner_htmls(self, elems):
        """
        Return a list of lists, where each outer list represents an element in
        ``elems``, and each inner list represents the ``innerHTML`` attribute
        of each item in the outer list.

        :param elems: A list of HTMLElements, such as the return value of
          :py:meth:`~.tbody2elemlist`
        :type elems: list
        :return: list of lists, rows to innerHTML of elements in each row
        :rtype: list
        """
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        return htmls

    def sort_trans_rows(self, rows):
        """
        Sort a list of transaction rows by date and then amount, to match up
        with the HTML in transactions table.

        :param rows: list of inner HTMLs, such as those returned by
          :py:meth:`~.inner_htmls`.
        :type rows: list
        :return: sorted rows
        :rtype: list
        """
        tmp_rows = []
        for row in rows:
            row[1] = Decimal(row[1].replace('$', ''))
            tmp_rows.append(row)
        ret = []
        for row in sorted(tmp_rows, key=lambda x: (x[0], x[1])):
            row[1] = fmt_currency(row[1])
            ret.append(row)
        return ret

    def try_click(self, driver, elem):  # noqa
        """
        Wrapper for recent Chrome Headless
        "Other element would receive the click" errors. Attempts to retry the
        click after a short wait if it throws that error.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param elem: element to click
        :type elem: selenium.webdriver.remote.webelement.WebElement
        """
        max_tries = 4
        for i in range(0, max_tries):
            try:
                elem.click()
                return
            except WebDriverException as ex:
                if 'Other element would receive the click' not in str(ex):
                    raise
                if i == max_tries - 1:
                    raise
                sleep(1.0)

    def try_click_and_get_modal(self, driver, elem_to_click, wait=True):
        """
        Combination of :py:meth:`~.try_click` and :py:meth:`~.get_modal_parts`
        to work around both the "Other element would receive the click" error
        and TimeoutExceptions waiting for the modal to be shown.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param elem_to_click: element to click
        :type elem_to_click: selenium.webdriver.remote.webelement.WebElement
        :param wait: whether or not to wait for presence of modalLabel
        :type wait: bool
        :return: 3-tuple of (modalDiv WebElement, modalLabel WebElement,
          modalBody WebElement)
        :rtype: tuple
        """
        max_tries = 4
        for i in range(0, max_tries):
            try:
                self.try_click(driver, elem_to_click)
                return self.get_modal_parts(driver, wait=wait)
            except (WebDriverException, TimeoutException):
                if i == max_tries - 1:
                    raise
                logger.error('ERROR: Unable to click link and get modal. '
                             'Trying again in 1s', exc_info=True)
                sleep(1.0)
        return None, None, None



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import datetime, date
from pytz import UTC
from decimal import Decimal
import re
import json
from time import sleep
from selenium.webdriver import ActionChains
import requests
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from biweeklybudget.utils import dtnow, fmt_currency
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models import *
from biweeklybudget.tests.sqlhelpers import restore_mysqldump
from biweeklybudget.tests.conftest import get_db_engine

dnow = dtnow()


def txn_div(id, dt, amt, acct_name, acct_id,
            budget_name, budget_id, desc, drop_div=''):
    """
    Return the HTML for a Transaction div.

    :param drop_div: contents of ``reconcile-drop-target`` div
    :type drop_div: str
    :return: HTML for Transaction reconcile div
    :rtype: str
    """
    s = '<div class="reconcile reconcile-trans ui-droppable" ' \
        'id="trans-%s" data-trans-id="%s" data-acct-id="%s" data-amt="%s">' % (
            id, id, acct_id, amt
        )
    s += '<div class="row">'
    s += '<div class="col-lg-3">%s</div>' % dt.strftime('%Y-%m-%d')
    s += '<div class="col-lg-3">%s</div>' % fmt_currency(amt)
    s += '<div class="col-lg-3"><strong>Acct:</strong> '
    s += '<span style="white-space: nowrap;">'
    s += '<a href="/accounts/%s">%s (%s)</a>' % (acct_id, acct_name, acct_id)
    s += '</span></div>'
    s += '<div class="col-lg-3"><strong>Budget:</strong> '
    s += '<span style="white-space: nowrap;">'
    if isinstance(budget_name, type([])) and budget_id is None:
        b1 = budget_name[0]
        b2 = budget_name[1]
        s += '<a href="/budgets/%s">%s (%s) (%s)</a><br>' % (
            b1[1], b1[0], b1[1], b1[2]
        )
        s += '<a href="/budgets/%s">%s (%s) (%s)</a>' % (
            b2[1], b2[0], b2[1], b2[2]
        )
    else:
        s += '<a href="/budgets/%s">%s (%s)</a>' % (
            budget_id, budget_name, budget_id
        )
    s += '</span></div>'
    s += '</div>'
    s += '<div class="row"><div class="col-lg-12">'
    if drop_div == '':
        s += '<div style="float: left;"><a href="javascript:transModal('
        s += '%s, function () { updateReconcileTrans(%s) })">Trans %s</a>' \
             ': %s</div>' % (id, id, id, desc)
        s += '<div style="float: right;" class="trans-no-ofx"><a ' \
             'href="javascript:transNoOfx(%s)" style="" title="Reconcile ' \
             'as never having a matching OFX Transaction">(no OFX)</a>' \
             '</div>' % id
    else:
        s += '<div style="float: left;"><span class="disabledEditLink">' \
             'Trans %s</span>: %s</div>' % (id, desc)
        s += '<div style="float: right;" class="trans-no-ofx"><a ' \
             'href="javascript:transNoOfx(%s)" style="display: none;" ' \
             'title="Reconcile as never having a matching OFX Transaction">' \
             '(no OFX)</a></div>' % id
    s += '</div></div>'
    s += '<div class="reconcile-drop-target">%s</div>' % drop_div
    s += '</div>'
    return s


def clean_fitid(fitid):
    return re.sub(r'\W', '', fitid)


def ofx_div(dt_posted, amt, acct_name, acct_id, trans_type, fitid, name,
            trans_id=None, ignored_reason=None):
    """
    Return the HTML for an OFXTransaction div.

    :param trans_id: if dropped on a Transaction div, the trans_id
    :type trans_id: int
    :return: HTML for OFXTransaction reconcile div
    :rtype: str
    """
    cfitid = clean_fitid(fitid)
    if int(amt) == amt:
        # JS doesn't put the trailing decimal on a ".0" number
        amt = int(amt)
    if trans_id is not None:
        classes = 'reconcile reconcile-ofx-dropped'
        _id = 'dropped-ofx-%s-%s' % (acct_id, cfitid)
    else:
        classes = 'reconcile reconcile-ofx ui-draggable ui-draggable-handle'
        _id = 'ofx-%s-%s' % (acct_id, cfitid)
    if ignored_reason is not None:
        classes += ' ui-draggable-disabled'
    s = '<div class="%s" id="%s" data-acct-id="%s" ' \
        'data-amt="%s" data-fitid="%s" style="">' % (
            classes, _id, acct_id, amt, fitid
        )
    if ignored_reason is not None:
        s += '<div class="row" id="ofx-%s-%s-noTrans" style=""><div ' \
             'class="col-lg-12"><p><strong>No Trans:</strong> %s</p></div>' \
             '</div>' % (acct_id, fitid, ignored_reason)
    s += '<div class="row">'
    s += '<div class="col-lg-3">%s</div>' % dt_posted.strftime('%Y-%m-%d')
    s += '<div class="col-lg-3">%s</div>' % fmt_currency(amt)
    s += '<div class="col-lg-3"><strong>Acct:</strong> '
    s += '<span style="white-space: nowrap;">'
    s += '<a href="/accounts/%s">%s (%s)</a>' % (acct_id, acct_name, acct_id)
    s += '</span></div>'
    s += '<div class="col-lg-3"><strong>Type:</strong> %s</div>' % trans_type
    s += '</div>'
    s += '<div class="row"><div class="col-lg-12">'
    s += '<div style="float: left;">'
    s += '<a href="javascript:ofxTransModal(%s, \'%s\', false)">%s</a>' % (
        acct_id, cfitid, fitid
    )
    s += ': %s' % name
    s += '</div>'
    if trans_id is None and ignored_reason is None:
        s += '<div style="float: right;" class="make-trans-link">'
        s += '<a href="javascript:makeTransFromOfx(%d, \'%s\')" ' \
             'title="Create Transaction from this OFX">(make trans)</a>' % (
                 acct_id, fitid
             )
        s += '<a href="javascript:ignoreOfxTrans(%d, \'%s\')" ' \
             'title="Ignore this OFX Transaction">(ignore)</a>' % (
                 acct_id, fitid
             )
        s += '</div>'
    elif ignored_reason is not None:
        s += '<div style="float: right;" class="make-trans-link">'
        s += '<a href="javascript:reconcileDoUnreconcileNoTrans(%s, \'%s\')">' \
             'Unignore</a>' % (
                 acct_id, fitid
             )
        s += '</div>'
    s += '</div>'
    s += '</div></div>'
    if trans_id is not None:
        return '<div style="text-align: right;"><a href="javascript:reconc' \
               'ileDoUnreconcile(%s, %s, \'%s\')">Unreconcile</a></div>%s' % (
                   trans_id, acct_id, fitid, s
               )
    return s


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestReconcile(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Reconcile Transactions - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


class ReconcileHelper(AcceptanceHelper):

    def get_reconciled(self, driver):
        """
        Execute javascript in the selenium browser to return the
        ``reconciled`` JavaScript object as a JSON string; deserialize the
        JSON and return the resulting dict.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :return: ``reconciled`` javascript variable from page
        :rtype: dict
        """
        script = 'return JSON.stringify(reconciled);'
        res = driver.execute_script(script)
        print("reconciled JSON: %s" % res)
        r = json.loads(res)
        return {int(x): r[x] for x in r}

    def test_00_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, get_db_engine(), with_data=False)

    def test_01_add_accounts(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank,
            re_interest_charge='^interest-charge',
            re_payment='^(payment|thank you)',
            re_late_fee='^Late Fee',
            re_other_fee='^re-other-fee'
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        b = Account(
            description='Second Bank Account',
            name='BankTwo',
            acct_type=AcctType.Bank,
            negate_ofx_amounts=True
        )
        testdb.add(b)
        b.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_02_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=Decimal('0.0'),
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=Decimal('500.00')
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=Decimal('0.00')
        ))
        testdb.flush()
        testdb.commit()

    def test_03_add_transactions(self, testdb):
        acct1 = testdb.query(Account).get(1)
        acct2 = testdb.query(Account).get(2)
        ibudget = testdb.query(Budget).get(1)
        e1budget = testdb.query(Budget).get(2)
        e2budget = testdb.query(Budget).get(3)
        # income - matches OFX1
        t1 = Transaction(
            date=date(2017, 4, 10),
            budget_amounts={ibudget: Decimal('-100.00')},
            budgeted_amount=Decimal('-100.00'),
            description='income',
            account=acct1,
            planned_budget=ibudget
        )
        testdb.add(t1)
        # one transaction - matches OFX2
        t2 = Transaction(
            date=date(2017, 4, 10),
            budget_amounts={e2budget: Decimal('250.00')},
            description='trans1',
            account=acct1
        )
        testdb.add(t2)
        # another transaction - matches OFX3
        st1 = ScheduledTransaction(
            amount=Decimal('500.0'),
            description='ST1',
            account=acct2,
            budget=e1budget,
            date=date(2017, 4, 10)
        )
        testdb.add(st1)
        t3 = Transaction(
            date=date(2017, 4, 11),
            budget_amounts={
                e1budget: Decimal('590.00'),
                e2budget: Decimal('10.00')
            },
            budgeted_amount=Decimal('500.0'),
            description='trans2',
            account=acct2,
            planned_budget=e1budget,
            scheduled_trans=st1
        )
        testdb.add(t3)
        # non-matched transaction
        t4 = Transaction(
            date=date(2017, 4, 14),
            budget_amounts={e2budget: Decimal('10.00')},
            description='trans3',
            account=acct2
        )
        testdb.add(t4)
        # matched ScheduledTransaction
        st2 = ScheduledTransaction(
            amount=Decimal('10.0'),
            description='ST2',
            account=acct1,
            budget=e2budget,
            day_of_month=13
        )
        testdb.add(st2)
        # pair that matches OFXT6 and OFXT7
        t5 = Transaction(
            date=date(2017, 4, 16),
            budget_amounts={e2budget: Decimal('25.00')},
            description='trans4',
            account=acct2
        )
        testdb.add(t5)
        t6 = Transaction(
            date=date(2017, 4, 17),
            budget_amounts={e2budget: Decimal('25.00')},
            description='trans5',
            account=acct2
        )
        testdb.add(t6)
        testdb.flush()
        testdb.commit()

    def test_04_add_ofx(self, testdb):
        acct1 = testdb.query(Account).get(1)
        acct2 = testdb.query(Account).get(2)
        stmt1 = OFXStatement(
            account=acct1,
            filename='a1.ofx',
            file_mtime=dnow,
            as_of=dnow,
            currency='USD',
            acctid='1',
            bankid='b1',
            routing_number='r1'
        )
        testdb.add(stmt1)
        stmt2 = OFXStatement(
            account=acct2,
            filename='a2.ofx',
            file_mtime=dnow,
            as_of=dnow,
            currency='USD',
            acctid='2',
            bankid='b2',
            routing_number='r2'
        )
        testdb.add(stmt2)
        ################
        # transactions #
        ################
        # matches Transaction 1
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFX1',
            trans_type='Deposit',
            date_posted=datetime(2017, 4, 10, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-100.0'),
            name='ofx1-income'
        ))
        # matches Transaction 2
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFX2',
            trans_type='Debit',
            date_posted=datetime(2017, 4, 11, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('250.0'),
            name='ofx2-trans1'
        ))
        # matches Transcation 3
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt2,
            fitid='OFX3',
            trans_type='Purchase',
            date_posted=datetime(2017, 4, 9, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-600.0'),
            name='ofx3-trans2-st1'
        ))
        # non-matched - have Transaction 4 same amt but wrong acct
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFXT4',
            trans_type='Purchase',
            date_posted=datetime(2017, 4, 14, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('10.0'),
            name='ofx4-st2'
        ))
        # matches ScheduledTransaction 2
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFXT5',
            trans_type='Foo',
            date_posted=datetime(2017, 4, 16, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('10.0'),
            name='ofx5'
        ))
        # pair of matched transactions - Transaction 4 and 5
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt2,
            fitid='OFXT6',
            trans_type='Foo',
            date_posted=datetime(2017, 4, 16, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-25.0'),
            name='ofx6'
        ))
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt2,
            fitid='OFXT7',
            trans_type='Foo',
            date_posted=datetime(2017, 4, 17, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-25.0'),
            name='ofx7'
        ))
        testdb.flush()
        testdb.commit()

    def test_05_add_reconciled(self, testdb):
        acct2 = testdb.query(Account).get(2)
        stmt2 = testdb.query(OFXStatement).get(2)
        e2budget = testdb.query(Budget).get(3)
        o = OFXTransaction(
            account=acct2,
            statement=stmt2,
            fitid='OFX8',
            trans_type='Purchase',
            date_posted=datetime(2017, 4, 17, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-600.0'),
            name='ofx8-trans4'
        )
        testdb.add(o)
        t = Transaction(
            date=date(2017, 4, 16),
            budget_amounts={e2budget: Decimal('600.00')},
            description='trans6',
            account=acct2
        )
        testdb.add(t)
        testdb.add(TxnReconcile(transaction=t, ofx_trans=o))
        testdb.flush()
        testdb.commit()


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestColumns(ReconcileHelper):

    def test_06_transactions(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                '1Income', 1,
                'income'
            ),
            txn_div(
                2,
                date(2017, 4, 10),
                250,
                'BankOne', 1,
                '3Periodic', 3,
                'trans1'
            ),
            txn_div(
                3,
                date(2017, 4, 11),
                600,
                'BankTwo', 2,
                [
                    ['2Periodic', 2, '$590.00'],
                    ['3Periodic', 3, '$10.00']
                ],
                None,
                'trans2'
            ),
            txn_div(
                4,
                date(2017, 4, 14),
                10,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans3'
            ),
            txn_div(
                5,
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans4'
            ),
            txn_div(
                6,
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans5'
            )
        ]
        assert actual_trans == expected_trans

    def test_07_verify_unreconciled_ofxtrans(self, testdb):
        assert len(OFXTransaction.unreconciled(testdb).all()) == 7

    def test_08_add_ignored_ofxtrans(self, testdb):
        """
        add OFXTransactions that shouldn't be listed because of their is_ fields
        """
        acct = testdb.query(Account).get(1)
        assert acct.re_interest_charge == '^interest-charge'
        assert acct.re_interest_paid is None
        assert acct.re_payment == '^(payment|thank you)'
        assert acct.re_late_fee == '^Late Fee'
        assert acct.re_other_fee == '^re-other-fee'
        stmt = testdb.query(OFXStatement).get(1)
        assert stmt.account_id == 1
        assert stmt.filename == 'a1.ofx'
        testdb.add(OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne.77.1',
            trans_type='Debit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('-20.00'),
            name='interest-charge BankOne.77.1'
        ))
        testdb.add(OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne.77.2',
            trans_type='Debit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('-20.00'),
            name='payment BankOne.77.2'
        ))
        testdb.add(OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne.77.3',
            trans_type='Debit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('-20.00'),
            name='thank you BankOne.77.3'
        ))
        testdb.add(OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne.77.4',
            trans_type='Debit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('-20.00'),
            name='Late Fee BankOne.77.4'
        ))
        testdb.add(OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne.77.5',
            trans_type='Debit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('-20.00'),
            name='re-other-fee BankOne.77.5'
        ))
        testdb.commit()

    def test_09_verify_unreconciled_ofxtrans(self, testdb):
        assert len(OFXTransaction.unreconciled(testdb).all()) == 7

    def test_10_ofxtrans(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1'
            ),
            ofx_div(
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                'Deposit',
                'OFX1',
                'ofx1-income'
            ),
            ofx_div(
                date(2017, 4, 11),
                250,
                'BankOne', 1,
                'Debit',
                'OFX2',
                'ofx2-trans1'
            ),
            ofx_div(
                date(2017, 4, 14),
                10,
                'BankOne', 1,
                'Purchase',
                'OFXT4',
                'ofx4-st2'
            ),
            ofx_div(
                date(2017, 4, 16),
                10,
                'BankOne', 1,
                'Foo',
                'OFXT5',
                'ofx5'
            ),
            ofx_div(
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT6',
                'ofx6'
            ),
            ofx_div(
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT7',
                'ofx7'
            )
        ]
        assert expected_ofx == actual_ofx


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestAccountReconcileFalse(ReconcileHelper):

    def test_06_transactions(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                '1Income', 1,
                'income'
            ),
            txn_div(
                2,
                date(2017, 4, 10),
                250,
                'BankOne', 1,
                '3Periodic', 3,
                'trans1'
            ),
            txn_div(
                3,
                date(2017, 4, 11),
                600,
                'BankTwo', 2,
                [
                    ['2Periodic', 2, '$590.00'],
                    ['3Periodic', 3, '$10.00']
                ],
                None,
                'trans2'
            ),
            txn_div(
                4,
                date(2017, 4, 14),
                10,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans3'
            ),
            txn_div(
                5,
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans4'
            ),
            txn_div(
                6,
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans5'
            )
        ]
        assert actual_trans == expected_trans

    def test_07_ofxtrans(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1'
            ),
            ofx_div(
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                'Deposit',
                'OFX1',
                'ofx1-income'
            ),
            ofx_div(
                date(2017, 4, 11),
                250,
                'BankOne', 1,
                'Debit',
                'OFX2',
                'ofx2-trans1'
            ),
            ofx_div(
                date(2017, 4, 14),
                10,
                'BankOne', 1,
                'Purchase',
                'OFXT4',
                'ofx4-st2'
            ),
            ofx_div(
                date(2017, 4, 16),
                10,
                'BankOne', 1,
                'Foo',
                'OFXT5',
                'ofx5'
            ),
            ofx_div(
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT6',
                'ofx6'
            ),
            ofx_div(
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT7',
                'ofx7'
            )
        ]
        assert expected_ofx == actual_ofx

    def test_08_set_do_not_reconcile(self, testdb):
        acct = testdb.query(Account).get(2)
        acct.reconcile_trans = False
        testdb.flush()
        testdb.commit()

    def test_09_transactions(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                '1Income', 1,
                'income'
            ),
            txn_div(
                2,
                date(2017, 4, 10),
                250,
                'BankOne', 1,
                '3Periodic', 3,
                'trans1'
            ),
        ]
        assert actual_trans == expected_trans

    def test_10_ofxtrans(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                'Deposit',
                'OFX1',
                'ofx1-income'
            ),
            ofx_div(
                date(2017, 4, 11),
                250,
                'BankOne', 1,
                'Debit',
                'OFX2',
                'ofx2-trans1'
            ),
            ofx_div(
                date(2017, 4, 14),
                10,
                'BankOne', 1,
                'Purchase',
                'OFXT4',
                'ofx4-st2'
            ),
            ofx_div(
                date(2017, 4, 16),
                10,
                'BankOne', 1,
                'Foo',
                'OFXT5',
                'ofx5'
            )
        ]
        assert expected_ofx == actual_ofx


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestTransactionEditModal(ReconcileHelper):

    def test_06_verify_db(self, testdb):
        t = testdb.query(Transaction).get(1)
        assert t is not None
        assert t.description == 'income'
        assert t.date == date(2017, 4, 10)
        assert t.actual_amount == Decimal('-100.00')
        assert t.account_id == 1
        assert t.planned_budget_id == 1
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 1
        assert t.budget_transactions[0].amount == Decimal('-100.00')

    def test_07_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        link = selenium.find_element(By.XPATH, '//a[text()="Trans 1"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 1'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '1'
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('-123.45')
        desc = body.find_element(By.ID, 'trans_frm_description')
        desc.send_keys('edited')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 1 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(t.get_attribute('outerHTML'))
            for t in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                Decimal('-123.45'),
                'BankOne', 1,
                '1Income', 1,
                'incomeedited'
            ),
            txn_div(
                2,
                date(2017, 4, 10),
                250,
                'BankOne', 1,
                '3Periodic', 3,
                'trans1'
            ),
            txn_div(
                3,
                date(2017, 4, 11),
                600,
                'BankTwo', 2,
                [
                    ['2Periodic', 2, '$590.00'],
                    ['3Periodic', 3, '$10.00']
                ],
                None,
                'trans2'
            ),
            txn_div(
                4,
                date(2017, 4, 14),
                10,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans3'
            ),
            txn_div(
                5,
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans4'
            ),
            txn_div(
                6,
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans5'
            )
        ]
        assert actual_trans == expected_trans


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestDragLimitations(ReconcileHelper):

    def test_06_success(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-2-OFX3')
        tgt = selenium.find_element(By.ID,
                                    'trans-3').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # ensure that the OFX div was hidden in the OFX column
        src = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert src.is_displayed() is False
        # ensure that the OFX div was placed in the drop target
        tgt = selenium.find_element(By.ID, 'trans-3')
        expected = txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2',
            drop_div=ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1',
                trans_id=3
            )
        )
        assert self.normalize_html(tgt.get_attribute('outerHTML')) == expected
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            3: [2, 'OFX3']
        }

    def test_07_already_has_ofx(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-2-OFXT6')
        src2 = selenium.find_element(By.ID, 'ofx-2-OFXT7')
        tgt = selenium.find_element(By.ID,
                                    'trans-5').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # ensure that the OFX div was hidden in the OFX column
        src = selenium.find_element(By.ID, 'ofx-2-OFXT6')
        assert src.is_displayed() is False
        # ensure that the OFX div was placed in the drop target
        tgt = selenium.find_element(By.ID, 'trans-5')
        expected = txn_div(
            5,
            date(2017, 4, 16),
            25,
            'BankTwo', 2,
            '3Periodic', 3,
            'trans4',
            drop_div=ofx_div(
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT6',
                'ofx6',
                trans_id=5
            )
        )
        assert self.normalize_html(tgt.get_attribute('outerHTML')) == expected
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            5: [2, 'OFXT6']
        }
        # get the innerHTML of both columns
        trans_div = selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML')
        # attempt to drag the other OFX
        chain = ActionChains(selenium)
        chain.drag_and_drop(src2, tgt).perform()
        # sleep a bit for the drag to stop
        sleep(1)
        # ensure both columns are still the same
        assert selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML') == trans_div
        assert selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML') == ofxtrans_div
        # ensure reconciled JS var is still the same
        assert self.get_reconciled(selenium) == {
            5: [2, 'OFXT6']
        }

    def test_08_wrong_account(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-1-OFXT4')
        tgt = selenium.find_element(By.ID,
                                    'trans-4').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # get the innerHTML of both columns
        trans_div = selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # sleep a bit for the drag to stop
        sleep(1)
        # ensure both columns are still the same
        assert selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML') == trans_div
        assert selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML') == ofxtrans_div
        # ensure reconciled JS var is still the same
        assert self.get_reconciled(selenium) == {}

    def test_09_wrong_amount(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-1-OFXT4')
        tgt = selenium.find_element(By.ID,
                                    'trans-1').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # get the innerHTML of both columns
        trans_div = selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # sleep a bit for the drag to stop
        sleep(1)
        # ensure both columns are still the same
        assert selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML') == trans_div
        assert selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML') == ofxtrans_div
        # ensure reconciled JS var is still the same
        assert self.get_reconciled(selenium) == {}

    def test_10_wrong_acct_and_amount(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-1-OFXT4')
        tgt = selenium.find_element(By.ID,
                                    'trans-3').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # get the innerHTML of both columns
        trans_div = selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # sleep a bit for the drag to stop
        sleep(1)
        # ensure both columns are still the same
        assert selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML') == trans_div
        assert selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML') == ofxtrans_div
        # ensure reconciled JS var is still the same
        assert self.get_reconciled(selenium) == {}

    def test_11_unreconcile(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/reconcile')
        src = selenium.find_element(By.ID, 'ofx-2-OFX3')
        tgt = selenium.find_element(By.ID,
                                    'trans-3').find_element(By.CLASS_NAME, 'reconcile-drop-target')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(src, tgt).perform()
        # ensure that the OFX div was hidden in the OFX column
        src = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert src.is_displayed() is False
        # ensure that the OFX div was placed in the drop target
        tgt = selenium.find_element(By.ID, 'trans-3')
        expected = txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2',
            drop_div=ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1',
                trans_id=3
            )
        )
        assert self.normalize_html(tgt.get_attribute('outerHTML')) == expected
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            3: [2, 'OFX3']
        }
        # unreconcile
        link = tgt.find_element(By.XPATH, '//a[text()="Unreconcile"]')
        link.click()
        src = selenium.find_element(By.ID, 'ofx-2-OFX3')
        tgt = selenium.find_element(By.ID, 'trans-3')
        assert src.is_displayed() is True
        assert self.normalize_html(src.get_attribute('outerHTML')) == ofx_div(
            date(2017, 4, 9),
            Decimal('600.00'),
            'BankTwo', 2,
            'Purchase',
            'OFX3',
            'ofx3-trans2-st1'
        )
        assert tgt.find_element(By.CLASS_NAME,
                                'reconcile-drop-target').get_attribute('innerHTML') == ''
        assert self.get_reconciled(selenium) == {}
        expected = txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2'
        )
        assert self.normalize_html(tgt.get_attribute('outerHTML')) == expected


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
@pytest.mark.skip  # See block comment in method below 2022-10-22
class TestDragAndDropReconcile(ReconcileHelper):

    def test_06_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert res[0].id == 1
        assert res[0].txn_id == 7
        assert res[0].ofx_account_id == 2
        assert res[0].ofx_fitid == 'OFX8'

    def test_07_drag_and_drop(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-2-OFX3'),
            selenium.find_element(By.ID,
                                  'trans-3'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        """
        IMPORTANT - 2022-10-22
        These tests used to work fine. They haven't been run in a couple of
        years. As of today, they're always failing with

            ElementNotInteractableException: Message: element not interactable:
            [object HTMLDivElement] has no size and location

        on the second drag-and-drop operation. No matter which divs are
        involved, or which order they're run in, the first drag-and-drop works
        fine and the second raises that exception.

        I'm currently using ChromeDriver 106 and Chrome 106. It's likely that
        the last time these tests were working was sometime in the era of
        ChromeDriver 79 or 81 and the coorresponding Chrome version.

        I've dug around on Google for about an hour and tried a bunch of things
        (like what's shown below) but just can't get this test to work. It's
        possible (likely?) that maybe I need some updates to webdriver /
        selenium or that I should try this test from a new standalone Python
        file outside of this project, with all the latest dependencies, and see
        if that works.

        But, given all of the other things I need to get done at much higher
        priority, I'm going to mark this class to be skipped and revisit it
        later.
        """
        # DEBUG
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located((By.ID, 'trans-1'))
        )
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'trans-1'))
        )
        sleep(2)
        chain.move_to_element(
            selenium.find_element(By.ID, 'trans-1')
        ).perform()
        # END DEBUG
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-1-OFX1'),
            selenium.find_element(By.ID,
                                  'trans-1'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-1-OFX2'),
            selenium.find_element(By.ID,
                                  'trans-2'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-2-OFXT6'),
            selenium.find_element(By.ID,
                                  'trans-5'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-2-OFXT7'),
            selenium.find_element(By.ID,
                                  'trans-6'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            3: [2, 'OFX3'],
            1: [1, 'OFX1'],
            2: [1, 'OFX2'],
            5: [2, 'OFXT6'],
            6: [2, 'OFXT7']
        }
        # click submit button
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Successfully reconciled 5 transactions'
        assert 'alert-success' in msg.get_attribute('class')

    def test_08_submit_with_nothing_reconciled(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        # get the innerHTML of both columns
        trans_div = selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML')
        # attempt to drag the other OFX
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        # ensure both columns are still the same
        assert selenium.find_element(By.ID, 'trans-panel').get_attribute(
            'innerHTML') == trans_div
        assert selenium.find_element(By.ID, 'ofx-panel').get_attribute(
            'innerHTML') == ofxtrans_div
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Warning: No reconciled transactions; ' \
                           'did not submit form.'
        assert 'alert-warning' in msg.get_attribute('class')


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
@pytest.mark.skip  # See block comment in method below 2022-10-22
class TestUIReconcileMulti(ReconcileHelper):
    """
    IMPORTANT - 2022-10-22
    These tests used to work fine. They haven't been run in a couple of years.
    As of today, they're always failing with

        ElementNotInteractableException: Message: element not interactable:
        [object HTMLDivElement] has no size and location

    on the second drag-and-drop operation. No matter which divs are involved,
    or which order they're run in, the first drag-and-drop works fine and
    the second raises that exception.

    I'm currently using ChromeDriver 106 and Chrome 106. It's likely that the
    last time these tests were working was sometime in the era of
    ChromeDriver 79 or 81 and the coorresponding Chrome version.

    I've dug around on Google for about an hour and tried a bunch of things
    (like what's shown below) but just can't get this test to work. It's
    possible (likely?) that maybe I need some updates to webdriver / selenium
    or that I should try this test from a new standalone Python file outside
    of this project, with all the latest dependencies, and see if that works.

    But, given all of the other things I need to get done at much higher
    priority, I'm going to mark this class to be skipped and revisit it later.
    """

    def test_06_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert res[0].id == 1
        assert res[0].txn_id == 7
        assert res[0].ofx_account_id == 2
        assert res[0].ofx_fitid == 'OFX8'

    def test_07_drag_and_drop(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-2-OFX3'),
            selenium.find_element(By.ID,
                                  'trans-3'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            3: [2, 'OFX3']
        }
        # click submit button
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Successfully reconciled 1 transactions'
        assert 'alert-success' in msg.get_attribute('class')
        # reconcile 2 more
        self.wait_for_id(selenium, 'ofx-1-OFX2')
        chain = ActionChains(selenium)
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-1-OFX1'),
            selenium.find_element(By.ID,
                                  'trans-1'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-1-OFX2'),
            selenium.find_element(By.ID,
                                  'trans-2'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            1: [1, 'OFX1'],
            2: [1, 'OFX2']
        }
        # click submit button
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Successfully reconciled 2 transactions'
        assert 'alert-success' in msg.get_attribute('class')

    def test_08_invalid_trans_id(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        assert self.get_reconciled(selenium) == {}
        script = 'reconciled[1234] = [4, "OFXNONE"];'
        selenium.execute_script(script)
        assert self.get_reconciled(selenium) == {
            1234: [4, "OFXNONE"]
        }
        # click submit button
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {
            1234: [4, "OFXNONE"]
        }
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Error 400: Invalid Transaction ID: 1234'
        assert 'alert-danger' in msg.get_attribute('class')


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestReconcileBackend(ReconcileHelper):

    def test_06_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert res[0].id == 1
        assert res[0].txn_id == 7
        assert res[0].ofx_account_id == 2
        assert res[0].ofx_fitid == 'OFX8'

    def test_07_success(self, base_url):
        res = requests.post(
            base_url + '/ajax/reconcile',
            json={'reconciled': {3: [2, 'OFX3']}, 'ofxIgnored': {}}
        )
        assert res.json() == {
            'success': True,
            'success_message': 'Successfully reconciled 1 transactions'
        }
        assert res.status_code == 200

    def test_08_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 2
        assert res[0].id == 1
        assert res[1].id == 2
        assert res[1].txn_id == 3
        assert res[1].ofx_account_id == 2
        assert res[1].ofx_fitid == 'OFX3'

    def test_09_invalid_trans(self, base_url, testdb):
        res = requests.post(
            base_url + '/ajax/reconcile',
            json={'reconciled': {32198: [2, 'OFX3']}, 'ofxIgnored': {}}
        )
        assert res.json() == {
            'success': False,
            'error_message': 'Invalid Transaction ID: 32198'
        }
        assert res.status_code == 400
        assert len(testdb.query(TxnReconcile).all()) == 2

    def test_10_invalid_ofx(self, base_url, testdb):
        res = requests.post(
            base_url + '/ajax/reconcile',
            json={'reconciled': {3: [2, 'OFX338ufd']}, 'ofxIgnored': {}}
        )
        assert res.json() == {
            'success': False,
            'error_message': "Invalid OFXTransaction: (2, 'OFX338ufd')"
        }
        assert res.status_code == 400
        assert len(testdb.query(TxnReconcile).all()) == 2

    def test_10_commit_exception(self, base_url):
        # already reconciled in test_07
        res = requests.post(
            base_url + '/ajax/reconcile',
            json={'reconciled': {3: [2, 'OFX3']}, 'ofxIgnored': {}}
        )
        j = res.json()
        assert sorted(j.keys()) == ['error_message', 'success']
        assert j['success'] is False
        assert j['error_message'].startswith('Exception committing reconcile')
        assert "Duplicate entry '3' for key " \
               "'uq_txn_reconciles_txn_id'" in j['error_message']
        assert res.status_code == 400

    def test_11_verify_db(self, testdb):
        testdb.expire_all()
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 2
        assert res[0].id == 1
        assert res[1].id == 2
        assert res[1].txn_id == 3
        assert res[1].ofx_account_id == 2
        assert res[1].ofx_fitid == 'OFX3'

    def test_12_reconcile_noOFX(self, base_url):
        res = requests.post(
            base_url + '/ajax/reconcile',
            json={'reconciled': {4: 'Foo Bar Baz'}, 'ofxIgnored': {}}
        )
        assert res.json() == {
            'success': True,
            'success_message': 'Successfully reconciled 1 transactions'
        }
        assert res.status_code == 200

    def test_13_verify_db(self, testdb):
        testdb.expire_all()
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 3
        assert res[2].txn_id == 4
        assert res[2].note == 'Foo Bar Baz'

    def test_14_verify_reconcile_modal(self, base_url, selenium, testdb):
        res = testdb.query(TxnReconcile).all()
        txn_id = res[-1].txn_id
        self.get(selenium, base_url + '/transactions')
        modal, title, body = self.try_click_and_get_modal(
            selenium, selenium.find_element(By.LINK_TEXT, 'Yes (%s)' % txn_id)
        )
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile %s' % txn_id
        assert 'Foo Bar Baz' in body.text
        assert body.find_elements(By.CLASS_NAME, 'col-lg-6')[1].text == \
            'No OFX Transaction'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestOFXMakeTransAndIgnore(AcceptanceHelper):

    def get_reconciled(self, driver):
        """
        Execute javascript in the selenium browser to return the
        ``reconciled`` JavaScript object as a JSON string; deserialize the
        JSON and return the resulting dict.

        :param driver: Selenium driver instance
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :return: ``reconciled`` javascript variable from page
        :rtype: dict
        """
        script = 'return JSON.stringify(reconciled);'
        res = driver.execute_script(script)
        print("reconciled JSON: %s" % res)
        r = json.loads(res)
        return {int(x): r[x] for x in r}

    def test_00_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, get_db_engine(), with_data=False)

    def test_01_add_accounts(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        b = Account(
            description='Second Bank Account',
            name='BankTwo',
            acct_type=AcctType.Bank,
            negate_ofx_amounts=True
        )
        testdb.add(b)
        b.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_02_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=Decimal('0.0'),
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=Decimal('500.00')
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=Decimal('0.00')
        ))
        testdb.flush()
        testdb.commit()

    def test_03_add_transactions(self, testdb):
        acct1 = testdb.query(Account).get(1)
        ibudget = testdb.query(Budget).get(1)
        # income - matches OFX1
        t = Transaction(
            date=date(2017, 4, 10),
            budget_amounts={ibudget: Decimal('-123.45')},
            budgeted_amount=Decimal('-123.45'),
            description='income',
            account=acct1,
            planned_budget=ibudget
        )
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_04_add_ofx(self, testdb):
        acct2 = testdb.query(Account).get(2)
        stmt1 = OFXStatement(
            account=acct2,
            filename='a2.ofx',
            file_mtime=dnow,
            as_of=dnow,
            currency='USD',
            acctid='2',
            bankid='b1',
            routing_number='r1'
        )
        testdb.add(stmt1)
        # matches Transaction 2
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt1,
            fitid='OFX2',
            trans_type='Debit',
            date_posted=datetime(2017, 4, 11, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('251.23'),
            name='ofx2-trans1'
        ))
        testdb.flush()
        testdb.commit()

    def test_06_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 0
        stmts = testdb.query(OFXStatement).all()
        assert len(stmts) == 1
        assert max([s.id for s in stmts]) == 1

    def test_07_verify_db_transaction(self, testdb):
        res = testdb.query(Transaction).all()
        assert len(res) == 1
        assert res[0].id == 1

    def test_08_trans_from_ofx(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX2')
        link = ofxdiv.find_element(By.XPATH, '//a[text()="(make trans)"]')
        # test the modal population
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add Transaction for OFX (2, OFX2)'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == date(
            2017, 4, 11).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '-251.23'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'ofx2-trans1'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwo']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '2'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', '1Income (i)'],
            ['2', '2Periodic'],
            ['3', '3Periodic']
        ]
        budget_sel.select_by_value('2')
        notes = selenium.find_element(By.ID, 'trans_frm_notes')
        assert notes.get_attribute(
            'value') == 'created from OFXTransaction(2, OFX2)'
        notes.send_keys('foo')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 2 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # ensure that the original OFX div is hidden
        assert selenium.find_element(By.ID, 'ofx-2-OFX2').is_displayed() is False
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            2: [2, 'OFX2']
        }
        # ensure that the Transaction was added, and the ofx moved to it
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(t.get_attribute('outerHTML'))
            for t in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                Decimal('-123.45'),
                'BankOne', 1,
                '1Income', 1,
                'income'
            ),
            txn_div(
                2,
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                '2Periodic', 2,
                'ofx2-trans1',
                drop_div=ofx_div(
                    date(2017, 4, 11),
                    Decimal('-251.23'),
                    'BankTwo', 2,
                    'Debit',
                    'OFX2',
                    'ofx2-trans1',
                    trans_id=2
                )
            )
        ]
        assert actual_trans == expected_trans
        # wait for submit button to be visible and clickable, and click it
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.invisibility_of_element_located((By.ID, 'modalDiv'))
        )
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'reconcile-submit'))
        )
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Successfully reconciled 1 transactions'
        assert 'alert-success' in msg.get_attribute('class')

    def test_09_verify_db_txnreconcile(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert res[0].id == 1
        assert res[0].txn_id == 2
        assert res[0].ofx_account_id == 2
        assert res[0].ofx_fitid == 'OFX2'

    def test_10_verify_db_transaction(self, testdb):
        res = testdb.query(Transaction).all()
        assert len(res) == 2
        assert res[1].id == 2
        assert res[1].account_id == 2
        assert res[1].date == date(2017, 4, 11)
        assert res[1].actual_amount == Decimal('-251.23')
        assert res[1].description == 'ofx2-trans1'
        assert res[1].notes == 'created from OFXTransaction(2, OFX2)foo'
        assert len(res[1].budget_transactions) == 1
        assert res[1].budget_transactions[0].budget_id == 2
        assert res[1].budget_transactions[0].amount == Decimal('-251.23')

    def test_30_add_ofx(self, testdb):
        acct2 = testdb.query(Account).get(2)
        stmt1 = testdb.query(OFXStatement).get(1)
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt1,
            fitid='OFX30',
            trans_type='Debit',
            date_posted=datetime(2017, 4, 11, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('251.23'),
            name='ofx2-trans30'
        ))
        testdb.add(OFXTransaction(
            account=acct2,
            statement=stmt1,
            fitid='OFX31',
            trans_type='Debit',
            date_posted=datetime(2017, 4, 10, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('192.86'),
            name='ofx2-trans31'
        ))
        testdb.flush()
        testdb.commit()

    def test_31_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert res[0].id == 1
        assert res[0].txn_id == 2
        assert res[0].ofx_account_id == 2
        assert res[0].ofx_fitid == 'OFX2'
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 1
        assert max([r.id for r in res]) == 1

    def test_32_verify_columns(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 10),
                Decimal('-192.86'),
                'BankTwo', 2,
                'Debit',
                'OFX31',
                'ofx2-trans31'
            ),
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30'
            )
        ]
        assert expected_ofx == actual_ofx

    def test_33_ignore_ofx(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX31')
        link = ofxdiv.find_element(By.XPATH, '//a[text()="(ignore)"]')
        # test the modal population
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Ignore OFXTransaction (2, "OFX31")'
        assert body.find_element(By.ID,
                                 'trans_frm_acct_id').get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_fitid').get_attribute('value') == 'OFX31'
        notes = selenium.find_element(By.ID, 'trans_frm_note')
        assert notes.get_attribute(
            'value') == ''
        notes.send_keys('My Note')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        sleep(1)
        # check that modal was hidden
        modal, title, body = self.get_modal_parts(selenium, wait=False)
        self.assert_modal_hidden(modal, title, body)
        # check that the JS variable has been updated
        res = selenium.execute_script('return JSON.stringify(ofxIgnored);')
        assert json.loads(res.strip()) == {'2%OFX31': 'My Note'}
        # check that the OFX div has been updated
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 10),
                Decimal('-192.86'),
                'BankTwo', 2,
                'Debit',
                'OFX31',
                'ofx2-trans31',
                ignored_reason='My Note'
            ),
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30'
            )
        ]
        assert expected_ofx == actual_ofx
        # check that the OFX div is no longer draggable
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX31')
        assert 'ui-draggable-disabled' in ofxdiv.get_attribute('class')
        # wait for submit button to be visible and clickable, and click it
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.invisibility_of_element_located((By.ID, 'modalDiv'))
        )
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'reconcile-submit'))
        )
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Successfully reconciled 1 transactions'
        assert 'alert-success' in msg.get_attribute('class')

    def test_34_verify_db(self, testdb):
        res = testdb.query(TxnReconcile).all()
        assert len(res) == 2
        assert max([r.id for r in res]) == 2
        vals = {r.id: r for r in res}
        tr = vals[2]
        assert tr.txn_id is None
        assert tr.ofx_account_id == 2
        assert tr.ofx_fitid == 'OFX31'
        assert tr.note == 'My Note'

    def test_35_verify_columns(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30'
            )
        ]
        assert expected_ofx == actual_ofx

    def test_36_ignore_and_unignore_ofx(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        # check that the OFX div has been updated
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30'
            )
        ]
        assert expected_ofx == actual_ofx
        # ignore
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX30')
        link = ofxdiv.find_element(By.XPATH, '//a[text()="(ignore)"]')
        # test the modal population
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Ignore OFXTransaction (2, "OFX30")'
        assert body.find_element(By.ID,
                                 'trans_frm_acct_id').get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_fitid').get_attribute('value') == 'OFX30'
        notes = selenium.find_element(By.ID, 'trans_frm_note')
        assert notes.get_attribute(
            'value') == ''
        notes.send_keys('My Note')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        sleep(1)
        # check that modal was hidden
        modal, title, body = self.get_modal_parts(selenium, wait=False)
        self.assert_modal_hidden(modal, title, body)
        # check that the JS variable has been updated
        res = selenium.execute_script('return JSON.stringify(ofxIgnored);')
        assert json.loads(res.strip()) == {'2%OFX30': 'My Note'}
        # check that the OFX div has been updated
        self.wait_for_jquery_done(selenium)
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30',
                ignored_reason='My Note'
            )
        ]
        assert expected_ofx == actual_ofx
        # check that the OFX div is no longer draggable
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX30')
        assert 'ui-draggable-disabled' in ofxdiv.get_attribute('class')
        # ok, now Unignore
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX30')
        link = ofxdiv.find_element(By.XPATH, '//a[text()="Unignore"]')
        link.click()
        # and test that everything was reverted...
        res = selenium.execute_script('return JSON.stringify(ofxIgnored);')
        assert json.loads(res.strip()) == {}
        # check that the OFX div has been updated
        self.wait_for_jquery_done(selenium)
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 11),
                Decimal('-251.23'),
                'BankTwo', 2,
                'Debit',
                'OFX30',
                'ofx2-trans30'
            )
        ]
        assert expected_ofx == actual_ofx
        # check that the OFX div is no longer draggable
        ofxdiv = selenium.find_element(By.ID, 'ofx-2-OFX30')
        assert 'ui-draggable-disabled' not in ofxdiv.get_attribute('class')


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestTransReconcileNoOfx(ReconcileHelper):

    def test_06_transactions_column(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        trans_div = selenium.find_element(By.ID, 'trans-panel')
        actual_trans = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in trans_div.find_elements(By.CLASS_NAME, 'reconcile-trans')
        ]
        expected_trans = [
            txn_div(
                1,
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                '1Income', 1,
                'income'
            ),
            txn_div(
                2,
                date(2017, 4, 10),
                250,
                'BankOne', 1,
                '3Periodic', 3,
                'trans1'
            ),
            txn_div(
                3,
                date(2017, 4, 11),
                600,
                'BankTwo', 2,
                [
                    ['2Periodic', 2, '$590.00'],
                    ['3Periodic', 3, '$10.00']
                ],
                None,
                'trans2'
            ),
            txn_div(
                4,
                date(2017, 4, 14),
                10,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans3'
            ),
            txn_div(
                5,
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans4'
            ),
            txn_div(
                6,
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                '3Periodic', 3,
                'trans5'
            )
        ]
        assert actual_trans == expected_trans

    def test_07_ofx_column(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        ofxtrans_div = selenium.find_element(By.ID, 'ofx-panel')
        actual_ofx = [
            self.normalize_html(x.get_attribute('outerHTML'))
            for x in ofxtrans_div.find_elements(By.CLASS_NAME, 'reconcile-ofx')
        ]
        expected_ofx = [
            ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1'
            ),
            ofx_div(
                date(2017, 4, 10),
                -100,
                'BankOne', 1,
                'Deposit',
                'OFX1',
                'ofx1-income'
            ),
            ofx_div(
                date(2017, 4, 11),
                250,
                'BankOne', 1,
                'Debit',
                'OFX2',
                'ofx2-trans1'
            ),
            ofx_div(
                date(2017, 4, 14),
                10,
                'BankOne', 1,
                'Purchase',
                'OFXT4',
                'ofx4-st2'
            ),
            ofx_div(
                date(2017, 4, 16),
                10,
                'BankOne', 1,
                'Foo',
                'OFXT5',
                'ofx5'
            ),
            ofx_div(
                date(2017, 4, 16),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT6',
                'ofx6'
            ),
            ofx_div(
                date(2017, 4, 17),
                25,
                'BankTwo', 2,
                'Foo',
                'OFXT7',
                'ofx7'
            )
        ]
        assert expected_ofx == actual_ofx

    def test_08_reconcile_unreconcile_noOFX_visible(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        # check Trans and OFX
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2'
        )
        ofx = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert self.normalize_html(ofx.get_attribute('outerHTML')) == ofx_div(
            date(2017, 4, 9),
            Decimal('600.00'),
            'BankTwo', 2,
            'Purchase',
            'OFX3',
            'ofx3-trans2-st1'
        )
        # drag and drop
        chain = ActionChains(selenium)
        chain.drag_and_drop(
            selenium.find_element(By.ID, 'ofx-2-OFX3'),
            selenium.find_element(By.ID,
                                  'trans-3'
                                  ).find_element(By.CLASS_NAME, 'reconcile-drop-target')
        ).perform()
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {
            3: [2, 'OFX3']
        }
        # check Trans and OFX
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2',
            drop_div=ofx_div(
                date(2017, 4, 9),
                Decimal('600.00'),
                'BankTwo', 2,
                'Purchase',
                'OFX3',
                'ofx3-trans2-st1',
                trans_id=3
            )
        )
        ofx = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert ofx.is_displayed() is False
        # unreconcile
        trans.find_element(By.XPATH, '//a[text()="Unreconcile"]').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        # check Trans and OFX
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2'
        )
        ofx = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert self.normalize_html(ofx.get_attribute('outerHTML')) == ofx_div(
            date(2017, 4, 9),
            Decimal('600.00'),
            'BankTwo', 2,
            'Purchase',
            'OFX3',
            'ofx3-trans2-st1'
        )
        # ensure the reconciled variable was updated
        assert self.get_reconciled(selenium) == {}
        # click submit button
        selenium.find_element(By.ID, 'reconcile-submit').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        assert self.get_reconciled(selenium) == {}
        msg = selenium.find_element(By.ID, 'reconcile-msg')
        assert msg.text == 'Warning: No reconciled transactions; ' \
            'did not submit form.'
        assert 'alert-warning' in msg.get_attribute('class')

    def test_09_reconcile_unreconcile_noOFX(self, base_url, selenium):
        self.get(selenium, base_url + '/reconcile')
        # check Trans and OFX
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2'
        )
        ofx = selenium.find_element(By.ID, 'ofx-2-OFX3')
        assert self.normalize_html(ofx.get_attribute('outerHTML')) == ofx_div(
            date(2017, 4, 9),
            Decimal('600.00'),
            'BankTwo', 2,
            'Purchase',
            'OFX3',
            'ofx3-trans2-st1'
        )
        assert self.get_reconciled(selenium) == {}
        # reconcile as noOFX
        modal, title, body = self.try_click_and_get_modal(
            selenium, trans.find_element(By.LINK_TEXT, '(no OFX)')
        )
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Reconcile Transaction 3 Without OFX'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '3'
        note = body.find_element(By.ID, 'trans_frm_note')
        note.clear()
        note.send_keys('My Trans Note')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        sleep(1)
        self.wait_for_jquery_done(selenium)
        # assert modal is hidden
        assert selenium.find_element(By.ID, 'modalDiv').is_displayed() is False
        # test trans div was updated
        noofx_div = '<div style="text-align: right;"><a href="' \
                    'javascript:reconcileDoUnreconcileNoOfx(3)">' \
                    'Unreconcile</a></div><div class="reconcile" ' \
                    'id="trans-3-noOFX" style=""><p><strong>No OFX:</strong>' \
                    ' My Trans Note</p></div>'
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2',
            drop_div=noofx_div
        )
        # test that reconciled var was updated
        assert self.get_reconciled(selenium) == {3: 'My Trans Note'}
        # unreconcile
        trans.find_element(By.LINK_TEXT, 'Unreconcile').click()
        trans = selenium.find_element(By.ID, 'trans-3')
        assert self.normalize_html(trans.get_attribute('outerHTML')) == txn_div(
            3,
            date(2017, 4, 11),
            600,
            'BankTwo', 2,
            [
                ['2Periodic', 2, '$590.00'],
                ['3Periodic', 3, '$10.00']
            ],
            None,
            'trans2'
        )
        assert self.get_reconciled(selenium) == {}



"""plaid models

Revision ID: f5a002127934
Revises: ac93b5c826a5
Create Date: 2020-04-10 21:11:42.366371

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql
from sqlalchemy_utc.sqltypes import UtcDateTime

# revision identifiers, used by Alembic.
revision = 'f5a002127934'
down_revision = 'ac93b5c826a5'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'plaid_items',
        sa.Column('item_id', sa.String(length=70), nullable=False),
        sa.Column('access_token', sa.String(length=70), nullable=True),
        sa.Column('institution_name', sa.String(length=100), nullable=True),
        sa.Column('institution_id', sa.String(length=50), nullable=True),
        sa.Column('last_updated', UtcDateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint('item_id', name=op.f('pk_plaid_items')),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'plaid_accounts',
        sa.Column('item_id', sa.String(length=70), nullable=False),
        sa.Column('account_id', sa.String(length=70), nullable=False),
        sa.Column('name', sa.String(length=100), nullable=True),
        sa.Column('mask', sa.String(length=20), nullable=True),
        sa.Column('account_type', sa.String(length=50), nullable=True),
        sa.Column('account_subtype', sa.String(length=50), nullable=True),
        sa.ForeignKeyConstraint(
            ['item_id'],
            ['plaid_items.item_id'],
            name=op.f('fk_plaid_accounts_item_id_plaid_items')
        ),
        sa.PrimaryKeyConstraint(
            'item_id', 'account_id', name=op.f('pk_plaid_accounts')
        ),
        mysql_engine='InnoDB'
    )
    op.add_column(
        'accounts',
        sa.Column('plaid_account_id', sa.String(length=70), nullable=True)
    )
    op.create_foreign_key(
        op.f('fk_accounts_plaid_item_id_plaid_accounts'),
        'accounts',
        'plaid_accounts',
        ['plaid_item_id', 'plaid_account_id'],
        ['item_id', 'account_id']
    )
    op.drop_column('accounts', 'plaid_token')


def downgrade():
    op.add_column(
        'accounts',
        sa.Column('plaid_token', mysql.VARCHAR(length=70), nullable=True)
    )
    op.drop_constraint(
        op.f('fk_accounts_plaid_item_id_plaid_accounts'),
        'accounts',
        type_='foreignkey'
    )
    op.drop_column('accounts', 'plaid_account_id')
    op.drop_table('plaid_accounts')
    op.drop_table('plaid_items')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import argparse
import logging
import atexit

from biweeklybudget.models.projects import Project, BoMItem
from biweeklybudget.db import init_db, db_session, cleanup_db
from biweeklybudget.cliutils import set_log_debug, set_log_info

from wishlist.core import Wishlist

logger = logging.getLogger(__name__)

# suppress requests logging
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)
requests_log.propagate = True


class WishlistToProject(object):

    def __init__(self):
        atexit.register(cleanup_db)
        init_db()

    def run(self):
        """
        Run the synchronization.

        :return: 2-tuple; count of successful syncs, total count of projects
          with associated wishlists
        :rtype: tuple
        """
        logger.debug('Beginning wishlist sync run')
        success = 0
        total = 0
        for list_url, proj in self._get_wishlist_projects():
            total += 1
            try:
                if self._do_project(list_url, proj):
                    success += 1
            except Exception:
                logger.error('Exception updating project %s with list %s',
                             proj, list_url, exc_info=True)
        return success, total

    def _do_project(self, list_url, project):
        """
        Update a project with information from its wishlist.

        :param list_url: Amazon wishlist URL
        :type list_url: str
        :param project: the project to update
        :type project: Project
        :return: whether or not the update was successful
        :rtype: bool
        """
        logger.debug('Handling project: %s', project)
        pitems = self._project_items(project)
        witems = self._wishlist_items(list_url)
        logger.debug('Project has %d items; wishlist has %d',
                     len(pitems), len(witems))
        for url, item in pitems.items():
            if url not in witems:
                logger.info(
                    '%s (%s) removed from amazon list; setting inactive',
                    item, url
                )
                item.is_active = False
                db_session.add(item)
        for url, item in witems.items():
            if url in pitems:
                bitem = pitems[url]
                logger.info('Updating %s from Amazon wishlist', bitem)
            else:
                bitem = BoMItem()
                bitem.project = project
                logger.info('Adding new BoMItem for wishlist %s', url)
            bitem.url = url
            bitem.is_active = True
            bitem.quantity = item['quantity']
            bitem.unit_cost = item['cost']
            bitem.name = item['name']
            db_session.add(bitem)
        logger.info('Committing changes for project %s url %s',
                    project, list_url)
        db_session.commit()
        return True

    def _project_items(self, proj):
        """
        Return all of the BoMItems for the specified project, as a dict of
        URL to BoMItem.

        :param proj: the project to get items for
        :type proj: Project
        :return: item URLs to BoMItems
        :rtype: dict
        """
        res = {}
        for i in db_session.query(BoMItem).filter(
            BoMItem.project.__eq__(proj)
        ).all():
            res[i.url] = i
        return res

    def _wishlist_items(self, list_url):
        """
        Get the items on the specified wishlist.

        :param list_url: wishlist URL
        :type list_url: str
        :return: dict of item URL to item details dict
        :rtype: dict
        """
        res = {}
        list_name = list_url.split('/')[-1]
        logger.debug('Getting wishlist items for wishlist: %s', list_name)
        items = [i for i in Wishlist(list_name)]
        logger.debug("Found %d items in list" % len(items))
        for item in items:
            item = item.jsonable()
            d = {'name': item['title'], 'url': item['url']}
            try:
                d['quantity'] = item['wanted_count']
            except Exception:
                d['quantity'] = 1
            if item['price'] > 0:
                d['cost'] = item['price']
            else:
                d['cost'] = item['marketplace_price']
            res[item['url']] = d
        return res

    def _get_wishlist_projects(self):
        """
        Find all projects with descriptions that begin with a wishlist URL.

        :return: list of (url, Project object) tuples
        :rtype: list
        """
        res = []
        logger.debug('Querying active projects for wishlist URLs')
        q = db_session.query(Project).filter(Project.is_active.__eq__(True))
        total_p = 0
        for p in q.all():
            total_p += 1
            if p.notes.strip() == '':
                continue
            u = p.notes.split(' ')[0]
            if self._url_is_wishlist(u):
                res.append((u, p))
        logger.info('Found %d of %d projects with wishlist URLs',
                    len(res), total_p)
        return res

    @staticmethod
    def _url_is_wishlist(url):
        """
        Determine if the given string or URL matches a wishlist.

        :param url: URL or string to test
        :type url: str
        :return: whether url is a wishlist URL
        :rtype: bool
        """
        return url.startswith('https://www.amazon.com/gp/registry/wishlist/')


def parse_args():
    p = argparse.ArgumentParser(
        description='Synchronize Amazon wishlists to projects, for projects '
                    'with Notes fields beginning with an Amazon public '
                    'wishlist URL'
    )
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    args = p.parse_args()
    return args


def main():
    global logger
    format = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=format)
    logger = logging.getLogger()

    args = parse_args()

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)
    if args.verbose <= 1:
        # if we're not in verbose mode, suppress routine logging for cron
        lgr = logging.getLogger('alembic')
        lgr.setLevel(logging.WARNING)
        lgr = logging.getLogger('biweeklybudget.db')
        lgr.setLevel(logging.WARNING)

    syncer = WishlistToProject()
    success, total = syncer.run()
    if success != total:
        logger.warning('Synced %d of %d project wishlists', success, total)
        raise SystemExit(1)
    raise SystemExit(0)


if __name__ == "__main__":
    main()



"""TxnReconcile txn_id not null

Revision ID: 5446559aecf7
Revises: f136418fc4ab
Create Date: 2017-04-26 15:57:04.536291

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = '5446559aecf7'
down_revision = 'f136418fc4ab'
branch_labels = None
depends_on = None


def upgrade():
    # When making changes to a column that has a foreign key, we need to drop
    # and then re-add the constraint
    op.execute('DELETE FROM txn_reconciles WHERE txn_id IS NULL;')
    op.execute('LOCK TABLES txn_reconciles WRITE, transactions WRITE;')
    op.drop_constraint('fk_txn_reconciles_txn_id_transactions',
                       'txn_reconciles', type_='foreignkey')
    op.alter_column(
        'txn_reconciles',
        'txn_id',
        existing_type=mysql.INTEGER(display_width=11),
        nullable=False
    )
    op.create_foreign_key('fk_txn_reconciles_txn_id_transactions',
                          'txn_reconciles', 'transactions', ['txn_id'], ['id'])
    op.execute('UNLOCK TABLES;')


def downgrade():
    # When making changes to a column that has a foreign key, we need to drop
    # and then re-add the constraint
    op.execute('LOCK TABLES txn_reconciles WRITE, transactions WRITE;')
    op.drop_constraint('fk_txn_reconciles_txn_id_transactions',
                       'txn_reconciles', type_='foreignkey')
    op.alter_column(
        'txn_reconciles',
        'txn_id',
        existing_type=mysql.INTEGER(display_width=11),
        nullable=True
    )
    op.create_foreign_key('fk_txn_reconciles_txn_id_transactions',
                          'txn_reconciles', 'transactions', ['txn_id'], ['id'])
    op.execute('UNLOCK TABLES;')



"""TxnReconcile allow txn_id to be null

Revision ID: 08b6358a04bf
Revises: 04e61490804b
Create Date: 2018-03-07 19:48:06.050926

"""
from alembic import op
from sqlalchemy.dialects import mysql
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = '08b6358a04bf'
down_revision = '04e61490804b'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        'txn_reconciles', 'txn_id',
        existing_type=mysql.INTEGER(display_width=11),
        nullable=True
    )


def downgrade():
    conn = op.get_bind()
    conn.execute(text("SET FOREIGN_KEY_CHECKS=0"))
    op.alter_column(
        'txn_reconciles', 'txn_id',
        existing_type=mysql.INTEGER(display_width=11),
        nullable=False
    )
    conn.execute(text("SET FOREIGN_KEY_CHECKS=1"))



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import logging
from sqlalchemy import create_engine
from alembic.command import upgrade, downgrade
from alembic.script import ScriptDirectory
from alembic.environment import EnvironmentContext  # noqa

from biweeklybudget.tests.migrations.alembic_helpers import (
    load_premigration_sql
)

logger = logging.getLogger(__name__)


@pytest.mark.migrations
class MigrationTest(object):
    """
    Class to assist with writing tests for Alembic migrations that change data.

    To implement:

    1. Subclass in a new module and mark with ``@pytest.mark.migrations``
    2. Set the ``migration_rev`` class attribute.
    3. Override ``data_setup``, ``verify_before`` and ``verify_after``.
    """

    migration_rev = None

    def data_setup(self, engine):
        """method to setup sample data in empty tables"""
        raise NotImplementedError("Must implement in subclass")

    def verify_before(self, engine):
        """method to verify data before forward migration, and after reverse"""
        raise NotImplementedError("Must implement in subclass")

    def verify_after(self, engine):
        """method to verify data after forward migration"""
        raise NotImplementedError("Must implement in subclass")

    def _migrate_up_to(self, rev, alembic_conf):
        """Apply all migrations up to but NOT INCLUDING rev"""
        upgrade(alembic_conf, rev)

    def _apply_migration(self, rev, alembic_conf):
        """Apply the migration specified by rev"""
        upgrade(alembic_conf, rev)

    def _reverse_migration(self, rev, alembic_conf):
        """Reverse the migration specified by rev"""
        downgrade(alembic_conf, rev)

    def _get_current_rev(self, alembic_conf, engine):
        script = ScriptDirectory.from_config(alembic_conf)
        with engine.connect() as conn:
            with EnvironmentContext(alembic_conf, script) as env_context:
                env_context.configure(conn, version_table="alembic_version")
                migration_context = env_context.get_context()
                revision = migration_context.get_current_revision()
        return revision

    def _rev_list(self, alembic_conf):
        script = ScriptDirectory.from_config(alembic_conf)
        revs = []
        for sc in script.walk_revisions(base="base", head="heads"):
            revs.append({
                'down_revision': sc.down_revision,
                'revision': sc.revision,
                'doc': sc.doc
            })
        revs.reverse()
        return revs

    def test_migration_roundtrip(self, uri_left, alembic_config_left):
        """DO NOT OVERRIDE. Method that runs migration tests."""
        alembic_config_left.set_section_option(
            'bwbTest', 'connstring', uri_left
        )
        load_premigration_sql(uri_left)
        engine = create_engine(uri_left)
        revs = self._rev_list(alembic_config_left)
        up_to_rev = None
        for rev in revs:
            if rev['revision'] == self.migration_rev:
                break
            up_to_rev = rev['revision']
        self._migrate_up_to(up_to_rev, alembic_config_left)
        logger.info('Migrated up to revision: %s', up_to_rev)
        self.data_setup(engine)
        self.verify_before(engine)
        self._apply_migration(self.migration_rev, alembic_config_left)
        logger.info(
            'Migrated up to revision: %s',
            self._get_current_rev(alembic_config_left, engine)
        )
        self.verify_after(engine)
        self._reverse_migration(up_to_rev, alembic_config_left)
        logger.info(
            'Migrated back to revision: %s',
            self._get_current_rev(alembic_config_left, engine)
        )
        self.verify_before(engine)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""
from datetime import date
from decimal import Decimal

from biweeklybudget.models.scheduled_transaction import ScheduledTransaction

import pytest


class TestScheduleType:

    def test_date(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            date=date(2025, 4, 15)
        )
        assert st.schedule_type == 'date'

    def test_monthly(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=15
        )
        assert st.schedule_type == 'monthly'

    def test_per_period(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            num_per_period=2
        )
        assert st.schedule_type == 'per period'

    def test_weekly(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=0  # Monday
        )
        assert st.schedule_type == 'weekly'

    def test_annual(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=4,
            annual_day=15
        )
        assert st.schedule_type == 'annual'

    def test_none(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test'
        )
        assert st.schedule_type is None


class TestRecurrenceStr:

    def test_date(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            date=date(2025, 4, 15)
        )
        assert st.recurrence_str == '2025-04-15'

    def test_monthly_first(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=1
        )
        assert st.recurrence_str == '1st'

    def test_monthly_second(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=2
        )
        assert st.recurrence_str == '2nd'

    def test_monthly_third(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=3
        )
        assert st.recurrence_str == '3rd'

    def test_monthly_fourth(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=4
        )
        assert st.recurrence_str == '4th'

    def test_per_period(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            num_per_period=3
        )
        assert st.recurrence_str == '3 per period'

    def test_weekly_monday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=0
        )
        assert st.recurrence_str == 'Every Monday'

    def test_weekly_tuesday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=1
        )
        assert st.recurrence_str == 'Every Tuesday'

    def test_weekly_wednesday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=2
        )
        assert st.recurrence_str == 'Every Wednesday'

    def test_weekly_thursday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=3
        )
        assert st.recurrence_str == 'Every Thursday'

    def test_weekly_friday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=4
        )
        assert st.recurrence_str == 'Every Friday'

    def test_weekly_saturday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=5
        )
        assert st.recurrence_str == 'Every Saturday'

    def test_weekly_sunday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=6
        )
        assert st.recurrence_str == 'Every Sunday'

    def test_annual_jan(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=1,
            annual_day=15
        )
        assert st.recurrence_str == 'Jan 15th'

    def test_annual_apr(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=4,
            annual_day=4
        )
        assert st.recurrence_str == 'Apr 4th'

    def test_annual_dec_first(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=12,
            annual_day=1
        )
        assert st.recurrence_str == 'Dec 1st'

    def test_annual_jul_fourth(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=7,
            annual_day=4
        )
        assert st.recurrence_str == 'Jul 4th'

    def test_annual_dec_25(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=12,
            annual_day=25
        )
        assert st.recurrence_str == 'Dec 25th'

    def test_none(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test'
        )
        assert st.recurrence_str is None


class TestValidators:

    def test_day_of_month_valid_min(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=1
        )
        assert st.day_of_month == 1

    def test_day_of_month_valid_max(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_month=28
        )
        assert st.day_of_month == 28

    def test_day_of_month_invalid_zero(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                day_of_month=0
            )

    def test_day_of_month_invalid_29(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                day_of_month=29
            )

    def test_num_per_period_valid(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            num_per_period=1
        )
        assert st.num_per_period == 1

    def test_num_per_period_invalid_zero(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                num_per_period=0
            )

    def test_day_of_week_valid_monday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=0
        )
        assert st.day_of_week == 0

    def test_day_of_week_valid_sunday(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            day_of_week=6
        )
        assert st.day_of_week == 6

    def test_day_of_week_invalid_negative(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                day_of_week=-1
            )

    def test_day_of_week_invalid_7(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                day_of_week=7
            )

    def test_annual_month_valid_jan(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=1,
            annual_day=15
        )
        assert st.annual_month == 1

    def test_annual_month_valid_dec(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=12,
            annual_day=15
        )
        assert st.annual_month == 12

    def test_annual_month_invalid_zero(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                annual_month=0,
                annual_day=15
            )

    def test_annual_month_invalid_13(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                annual_month=13,
                annual_day=15
            )

    def test_annual_day_valid_min(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=1,
            annual_day=1
        )
        assert st.annual_day == 1

    def test_annual_day_valid_max(self):
        st = ScheduledTransaction(
            amount=Decimal('100.00'),
            description='Test',
            annual_month=1,
            annual_day=31
        )
        assert st.annual_day == 31

    def test_annual_day_invalid_zero(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                annual_month=1,
                annual_day=0
            )

    def test_annual_day_invalid_32(self):
        with pytest.raises(AssertionError):
            ScheduledTransaction(
                amount=Decimal('100.00'),
                description='Test',
                annual_month=1,
                annual_day=32
            )
